#!/usr/bin/perl
# divide_edges.pl: Mark Hasegawa-Johnson, March 2004
# Illinois Speech and Language Engineering
#
# Usage text (printed if divide_edges.pl is called with no arguments):
$USAGE=<<ENDUSAGE;

divide_edges.pl: Divide every edge into equal-sized sub-edges.

Usage:
 divide_edges.pl divider script.scp

The script file should contain input & output filenames in pairs,
i.e., "input1.lat output1.lat input2.lat output2.lat ...".  Each input
lattice file should contain a word lattice for exactly one utterance.
Every edge in the word lattice should have a p= tag, specifying the
pronunciation of that edge.  The pronunciation of each edge is split
at all occurrences of the regexp given by "divider", resulting in a
length-N array of symbols on each edge.  The temporal length of the
edge is then divided into N equal-sized units.  

Generated sub-edges have the following properties:

 1. S= and E= tags are determined by dividing up the temporal duration
 of the original edge into equal-length segments.

 2. p= tag is determined by dividing the pronunciation of the original
 edge at boundaries specified by the regexp "divider".

 3. W= tag is set to ${word}_${n}, where $word is the word
 transcription of the original edge, and 1\le $n\le number of subedges.

 4. a= tag and l= tag are set equal to 1/N times the a= tag and l= tag
 of the original edge.

 5. All other tags are simply copied from the input edge to the output
 edge.

ENDUSAGE

####################################################################
# GLOBAL VARIABLES

# Math functions
sub abs { my($x)=shift; ($x>0) ? $x : -$x ; }
sub round { $_[0] =~ /([^\.]*)(.*)/; ($2 >= 0.5) ? $1+1 : $1; }

###################################################
# Subroutine to read an HTK lattice file
#
# Syntax: ($nodes, $edges, $header) = read_HTK_lattice(\*FILEHANDLE)
#
# $nodes is an arrayref of nodes
#    each node is an arrayref containing the following data:
#    [ back, forw, time ]
#    where forw is an arrayref of indices of forward edges,
#    back is an arrayref of indices of backward edges.
#    time is the time-alignment of this node
# $edges is an arrayref of edges
#    each edge, $e, is a hashref
#    tags on input edge are interpreted as key=val pairs,
#    and loaded as $$e{$key}=$val
#    For example, $$e{'W'} is set to the word label
#    
# $header is a hashref containing the header information
#
sub read_HTK_lattice {
  my($fh) = $_[0];
  my(@nodes) = ();
  my(@edges) = ();
  my(%header) = ();
  my($line);

  # Read lines of the file
  while($line = <$fh>) {
    # Node number given by I=
    if($line =~ /I=(\d+)/) {
      my($nodenum) = $1;
      unless(exists($nodes[$nodenum])) { $nodes[$nodenum] = [ [],[],0 ]; }
      if($line =~ /t=([\.\d]+)/) { $nodes[$nodenum][2] = $1; }
    }
    # Edge number given by J=
    elsif($line =~ s/J=(\d+)//) {
      my($edgenum) = $1;
      my(%edge)=();

      # Look for the startnode marker
      if($line =~ s/S=(\d+)//) { 
	$edge{'S'} = $1;
	# Tell the start node that I am one of its forward edges
	push(@{$nodes[$1][1]}, $edgenum);
      }

      # Look for the endnode marker
      if($line =~ s/E=(\d+)//) { 
	$edge{'E'} = $1; 
	# Tell the end node that I am one of its backward edges
	push(@{$nodes[$1][0]}, $edgenum);
      }

      # Read all other field-value pairs
      while($line =~ s/(\S+)=(\S+)//) { $edge{$1}=$2; }
      # Store the edge
      $edges[$edgenum] = \%edge;
    }
    # Otherwise, try to read header information
    else {
      while ($line =~ s/(\S+)=(\S+)//) { $header{$1} = $2; }
    }    
  }

  return(\@nodes, \@edges, \%header);
}

###################################################
# Subroutine to write an HTK lattice file
#
# Syntax: ($N, $L) = write_HTK_lattice($nodes, $edges, $header, \*FILEHANDLE)
#
# $nodes is an arrayref of nodes
# $edges is an arrayref of edges
# $header is a hashref containing HTK lattice header information
# FILEHANDLE is the opened output file
#
# $N is the number of nodes actually written to the file
# $L is the number of edges actually written to the file
#
sub write_HTK_lattice {
  my($nodes) = $_[0];
  my($edges) = $_[1];
  my($header) = $_[2];
  my($fh) = $_[3];
  my($N) = 0;
  my($L) = 0;

  # Print header information
  print $fh "VERSION=1.0\n";
  print $fh "UTTERANCE=",$$header{'UTTERANCE'},"\n";
  print $fh "lmname=",$$header{'lmname'},"\n";
  print $fh "lmscale=",$$header{'lmscale'}," ";
  print $fh "wdpenalty=",$$header{'wdpenalty'},"\n";
  print $fh "vocab=",$$header{'vocab'},"\n";
  print $fh sprintf("N=%-5d L=%-7d\n",0+@{$nodes},0+@{$edges});

  # Print nodes
  foreach $N (0..$#{$nodes}) {
    print $fh sprintf("I=%-5d ",$N);
    if (exists($$nodes[$N][2])) 
      { print $fh sprintf("t=%4.2f",$$nodes[$N][2]); }
    print $fh "\n";
  }
  # Print edges
  foreach $L (0..$#{$edges}) {
    my($e)=$$edges[$L];

    # Print the standard key-value pairs
    print $fh sprintf("J=%-7d S=%-5d E=%-5d W=%-20s ",$L, 
		      $$e{'S'}, $$e{'E'}, $$e{'W'});
    # Print all other key-value pairs
    foreach (keys %$e) {
      unless(/^[SEW]$/) { print $fh sprintf("%s=%s ", $_, $$e{$_}); }
    }
    # Print newline
    print $fh "\n";
  }
  return($N,$L);
}

###################################################
#
# ($nodes, $edges) = split_edge($ei, $nodes, $edges)
#
# $nodes is an arrayref of nodes
# $edges is an arrayref of edges
# $edges[$ei] is the edge to be split
# New value of arrayrefs $nodes and $edges get returned 
#  -- note that the new values are not sorted in time!!!
#
sub split_edge {
  my($ei) = $_[0];
  my($nodes) = $_[1];
  my($edges) = $_[2];
  my($m); my($n);

  # Split the pronunciation at $DIVIDER
  my(@units) = split(/$DIVIDER/, $$edges[$ei]{'p'});

  # If there are 1 or fewer units, quit
  if(@units <= 1) { return($nodes, $edges); }
  
  # Compute the indices of all subedges and of all nodes
  my(@subedges) = ( $ei, ($#$edges+1)..($#$edges+$#units) );
  my(@subnodes) = ( $$edges[$ei]{'S'}, 
		    ($#$nodes+1)..($#$nodes+$#units),
		    $$edges[$ei]{'E'} );

  # Duration of each sub-edge = time of end node - time of start node
  my($starttime) = $$nodes[$subnodes[0]][2];
  my($endtime) = $$nodes[$subnodes[$#subnodes]][2];
  my($d) = ($endtime-$starttime)/@units;
  
  # Create new edges, fill with content equal to $$edges[$ei]
  foreach $n (1..$#units) { 
    $$edges[$subedges[$n]] = {};
    foreach $key (keys %{$$edges[$ei]}) 
      { $$edges[$subedges[$n]]{$key} = $$edges[$ei]{$key}; }
  }
  # Modify new edges: word, start node, end node, pronunciation
  foreach $n (0..$#units) {
    $$edges[$subedges[$n]]{'W'} .= '_' . ($n+1);
    $$edges[$subedges[$n]]{'S'} = $subnodes[$n];
    $$edges[$subedges[$n]]{'E'} = $subnodes[$n+1];
    $$edges[$subedges[$n]]{'p'} = $units[$n];
    $$edges[$subedges[$n]]{'a'} /= (1+$#units);
    $$edges[$subedges[$n]]{'l'} /= (1+$#units);
  }
  
  # Create nodes
  foreach $n (1..$#units) { 
    $$nodes[$subnodes[$n]] = [ [$subedges[$n-1]], [$subedges[$n]], 
			       $starttime+($n*$d) ];
  }
  # Finally, check the ending node:
  #   if its back nodelist refers to $ei, change it to $subedges[$#subedges]
  foreach $n (0..$#{$$nodes[$subnodes[$#subnodes]][0]}) {
    if($$nodes[$subnodes[$#subnodes]][0][$n] == $ei)
      { $$nodes[$subnodes[$#subnodes]][0][$n] = $subedges[$#subedges]; }
  }

  # Return final result
  return($nodes, $edges);
}

##################################################
# Main program

if (@ARGV<2) { print $USAGE; exit; }

# Read in the script file
$arg=$ARGV[1];
open(F,$arg) || die "Unable to read from script file $arg: $!\n";
$script = ''; while (<F>) { $script .= $_; }
close(F);

# Read in the divider
$DIVIDER=$ARGV[0];

# Main loop: process each pair of lattice files in the script
while($script =~ s/^\s*(\S+)\s+(\S+)//) {
  my($inlat) = $1; my($outlat) = $2;

  # Read the input lattice
  unless(open(LATFILE, $inlat)) {
    print STDERR "Unable to read from lattice $inlat: $!\n";
    next;
  }
  my(@graph) = read_HTK_lattice(\*LATFILE);
  close(LATFILE);

  # Split each of the edges
  my($ei)=0;
  foreach $ei (0..$#{$graph[1]}) {
    # Split this edge.  Append extra nodes & edges to $graph[0] and $graph[1]
    ($graph[0], $graph[1]) = split_edge($ei, $graph[0], $graph[1]); 
  }

  # Create a mapping from node time to current node number
  #   space-nodenum must be appended to time, so different nodes
  #   with the same time alignment can remain distinct
  my(%time2oldnode) = ();  my($n)=0;
  foreach $n (0..$#{$graph[0]}) 
    { $time2oldnode{$graph[0][$n][2].' '.$n} = $n; }

  # new2old: list the oldnodes in order of sorted node time,
  #   so that $new2old[$new]==$old, $old2new[$old]==$new
  my(@old2new) = ();
  my(@new2old) = ();
  foreach $t (sort {$a <=> $b} keys %time2oldnode) 
    { push(@new2old,$time2oldnode{$t}); }
  foreach $n (0..$#new2old) { $old2new[$new2old[$n]] = $n; }
    
  # Create a new graph, with same header as original
  my(@newgraph) = ( [], [], $graph[2] );
  # Copy the node arrayrefs, in the new order
  foreach $n (0..$#new2old) {
    $newgraph[0][$n] = $graph[0][$new2old[$n]];

    # Copy all edges that start at this node
    foreach $ei (@{$newgraph[0][$n][1]}) {
      # Update start and end node numbers to the new node numbers
      $graph[1][$ei]{'S'} = $old2new[$graph[1][$ei]{'S'}];
      $graph[1][$ei]{'E'} = $old2new[$graph[1][$ei]{'E'}];
      # Add this edge to the new graph
      push(@{$newgraph[1]}, $graph[1][$ei]);
    }
  }

  # Write the modified lattice
  unless(open(LAT_OUT, ">$outlat")) {
    print STDERR "Unable to write lattice out to $outlat: $!\n";
    next;
  }
  write_HTK_lattice(@newgraph, \*LAT_OUT);
  close(LAT_OUT);

  # Write summary to STDERR
  $nn=0+@{$newgraph[0]};
  $ne=0+@{$newgraph[1]};
  print STDERR "$inlat -> $outlat, $nn nodes, $ne edges\n";
}
