#!/usr/bin/perl
# dp_pinch.pl: Mark Hasegawa-Johnson, March 2004
# Illinois Speech and Language Engineering
#
# Usage text (printed if dp_pinch.pl is called with no arguments):
$USAGE=<<ENDUSAGE;

dp_pinch.pl: Pinch a lattice as much as can be done without changing
the number of edges in any existing path through the lattice.

Usage:
 dp_pinch.pl script.scp

DP_pinch.pl tries to merge all nodes in the lattice with nodes in the
MAP path, but only to the extent that this operation can be performed
without changing the number of edges in any existing path through the
lattice.  The MAP path __must be marked__: every edge in the MAP path
must be tagged with the tag "n=1" (see augment_lattice.pl).

The entire lattice is searched using a dynamic programming algorithm.
Every node in the lattice is either aligned with one of the nodes in
the MAP path, or marked as an "insertion."  If any edge skips over a
node in the MAP path, the MAP node is said to be "deleted" along that
path.  The cost of aligning a non-MAP node with any MAP node is equal
to the square of the distance between their node times.  The cost of a
node "insertion" or a node "deletion" is much, much higher than any
temporal misalignment.  The result is that no path through the
original lattice with the same number of nodes as the MAP path should
ever have any inserted or deleted nodes: nodes are only inserted or
deleted when there is no other way to achieve alignment.  After the DP
search, all aligned nodes are merged.

The script file should contain input & output filenames in pairs,
i.e., "input1.lat output1.lat input2.lat output2.lat ...".  Each input
lattice file should contain a word lattice for exactly one utterance.

ENDUSAGE

####################################################################
# GLOBAL VARIABLES

# Conversion rates for various file formats
$MLFPERIOD = 0.0000001;

# Insertion & deletion penalties -- larger than (t1-t2)^2 for any t1, t2
$NODE_INSERTION_COST=1e10;
$NODE_DELETION_COST=1e10;

# Insertion alignment threshold -- align two insertion nodes if
#   absolute difference in time is less than about 50ms
$INSERTION_ALIGNMENT_THRESHOLD=0.05;

# Math functions
sub abs { my($x)=shift; ($x>0) ? $x : -$x ; }
sub round { $_[0] =~ /([^\.]*)(.*)/; ($2 >= 0.5) ? $1+1 : $1; }

###################################################
# Subroutine to read an HTK lattice file
#
# Syntax: ($nodes, $edges, $header) = read_HTK_lattice(\*FILEHANDLE)
#
# $nodes is an arrayref of nodes
#    each node is an arrayref containing the following data:
#    [ back, forw, time ]
#    where forw is an arrayref of indices of forward edges,
#    back is an arrayref of indices of backward edges.
#    time is the time-alignment of this node
# $edges is an arrayref of edges
#    each edge, $e, is a hashref
#    tags on input edge are interpreted as key=val pairs,
#    and loaded as $$e{$key}=$val
#    For example, $$e{'W'} is set to the word label
#    
# $header is a hashref containing the header information
#
sub read_HTK_lattice {
  my($fh) = $_[0];
  my(@nodes) = ();
  my(@edges) = ();
  my(%header) = ();
  my($line);

  # Read lines of the file
  while($line = <$fh>) {
    # Node number given by I=
    if($line =~ /I=(\d+)/) {
      my($nodenum) = $1;
      unless(exists($nodes[$nodenum])) { $nodes[$nodenum] = [ [],[],0 ]; }
      if($line =~ /t=([\.\d]+)/) { $nodes[$nodenum][2] = $1; }
    }
    # Edge number given by J=
    elsif($line =~ s/J=(\d+)//) {
      my($edgenum) = $1;
      my(%edge)=();

      # Look for the startnode marker
      if($line =~ s/S=(\d+)//) 
      { 
	$edge{'S'} = $1;
	# Tell the start node that I am one of its forward edges
	push(@{$nodes[$1][1]}, $edgenum);
      }

      # Look for the endnode marker
      if($line =~ s/E=(\d+)//) 
      { 
	$edge{'E'} = $1; 
	# Tell the end node that I am one of its backward edges
	push(@{$nodes[$1][0]}, $edgenum);
      }

      # Read all other field-value pairs
      while($line =~ s/(\S+)=(\S+)//) { $edge{$1}=$2; }
      # Store the edge
      $edges[$edgenum] = \%edge;
    }
    # Otherwise, try to read header information
    else {
      while ($line =~ s/(\S+)=(\S+)//) { $header{$1} = $2; }
    }    
  }

  return(\@nodes, \@edges, \%header);
}

###################################################
# Subroutine to write an HTK lattice file
#
# Syntax: ($N, $L) = write_HTK_lattice($nodes, $edges, $header, \*FILEHANDLE)
#
# $nodes is an arrayref of nodes
# $edges is an arrayref of edges
# $header is a hashref containing HTK lattice header information
# FILEHANDLE is the opened output file
#
# $N is the number of nodes actually written to the file
# $L is the number of edges actually written to the file
#
sub write_HTK_lattice {
  my($nodes) = $_[0];
  my($edges) = $_[1];
  my($header) = $_[2];
  my($fh) = $_[3];
  my($N) = 0;
  my($L) = 0;

  # Print header information
  print $fh "VERSION=1.0\n";
  print $fh "UTTERANCE=",$$header{'UTTERANCE'},"\n";
  print $fh "lmname=",$$header{'lmname'},"\n";
  print $fh "lmscale=",$$header{'lmscale'}," ";
  print $fh "wdpenalty=",$$header{'wdpenalty'},"\n";
  print $fh "vocab=",$$header{'vocab'},"\n";
  print $fh sprintf("N=%-5d L=%-7d\n",0+@{$nodes},0+@{$edges});

  # Print nodes
  foreach $N (0..$#{$nodes}) {
    print $fh sprintf("I=%-5d ",$N);
    if (exists($$nodes[$N][2])) 
      { print $fh sprintf("t=%4.2f",$$nodes[$N][2]); }
    print $fh "\n";
  }
  # Print edges
  foreach $L (0..$#{$edges}) {
    my($e)=$$edges[$L];

    # Print the standard key-value pairs
    print $fh sprintf("J=%-7d S=%-5d E=%-5d W=%-20s ",$L, 
		      $$e{'S'}, $$e{'E'}, $$e{'W'});
    # Print all other key-value pairs
    foreach (keys %$e) {
      unless(/^[SEW]$/) { print $fh sprintf("%s=%s ", $_, $$e{$_}); }
    }
    # Print newline
    print $fh "\n";
  }
  return($N,$L);
}

###################################################
# dp_align - align nodes to times 
#
# ( $align, $mapnodes ) = dp_align($nodes, $edges, $mapnodes)
#
# $nodes - arrayref of nodes
# $edges - arrayref of edges
# $mapnodes - arrayref containing indices of nodes in the MAP path
#    The $mapnodes array should be sorted when this function begins.
#    This function adds new nodes to $mapnodes, and the new nodes
#      are NOT SORTED!
#
# This search algorithm assumes that the graph must start at $$nodes[0]
# and end at $$nodes[$#$nodes].  Furthermore, we assume that
# $$mapnodes[0]==0, that is, that the 0th node is the first node in
# the MAP path.
#
# Other nodes may align with a mapnode, or may be insertions.
# Other mapnodes may align with a node, or may be deletions.
# The total cost of aligning node $i at or after time $n is the sum
#  of the cost of the path from (0,0) to ($n,$i), plus the cost of the 
#  path from ($n,$i) to ($#$times,$#$nodes).
#
# $align - arrayref, of the same length as $nodes, 
#   $$nodes[$n]==$m means that $$nodes[$n] should be merged with $$nodes[$m]
#   Note that it's possible for $$nodes[$n] to equal $n
#
sub dp_align {
  my($nodes) = $_[0];
  my($edges) = $_[1];
  my($mapnodes) = $_[2];
  my(@sub_cost) = ();
  my($n); my($i);
  if(@$mapnodes < 1) { 
    print STDERR "dp_align called with empty mapnodes array\n";
    return([], $mapnodes); 
  }
  if(@$nodes < 1) { 
    print STDERR "dp_align called with empty nodes array\n";
    return([], $mapnodes); 
  }
  if(@$edges < 1) { 
    print STDERR "dp_align called with empty edges array\n";
    return([], $mapnodes); 
  }

  # Find the times of mapnodes
  my(@times) = ();
  foreach $n (0..$#$mapnodes) { $times[$n]=$$nodes[$$mapnodes[$n]][2]; }

  # Compute substitution costs for each node at each level
  foreach $n (0..$#times) { foreach $i (0..$#{$nodes}) {
    # Substitution cost = (time_diff)^2
    $sub_cost[$n][$i] = $times[$n]-$$nodes[$i][2];
    $sub_cost[$n][$i] = $sub_cost[$n][$i] * $sub_cost[$n][$i];
  }}

  #######
  # Forward cost computation
  # Initialize: at level 0, cost to each node determined by 
  #    number of insertions, plus alignment of $$nodes[0]
  my(@E1) = ();   $E1[0][0] = $sub_cost[0][0];  
  my(@bl) = ();   $bl[0][0] = -1;
  my(@be) = ();   $be[0][0] = -1;
  foreach $i (1..$#{$nodes}) {
    $E1[0][$i] = $sub_cost[0][0] + 100*$NODE_INSERTION_COST;
    foreach $ei (@{$$nodes[$i][0]}) {
      if($E1[0][$$edges[$ei]{'S'}] < $E1[0][$i] - $NODE_INSERTION_COST) {
	$E1[0][$i] = $E1[0][$$edges[$ei]{'S'}] + $NODE_INSERTION_COST; 
	$be[0][$i] = $ei; $bl[0][$i] = 0;
      }
    }
  }

  # Now compute minimum-cost path at each higher level
  foreach $n (1..$#times) {
    foreach $i (0..$#{$nodes}) {
      # $E1[$n][$i] is no larger than $E1[$n-1][$i] + one deletion
      $E1[$n][$i] = $E1[$n-1][$i] + $NODE_DELETION_COST;
      $be[$n][$i] = -1; $bl[$n][$i] = $n-1; 

      # Search through all edges going backward from this node
      foreach $ei (@{$$nodes[$i][0]}) {
	# First, look back and down: is sub_cost less than current best?
	if($E1[$n-1][$$edges[$ei]{'S'}] < $E1[$n][$i] - $sub_cost[$n][$i]) {
	  $E1[$n][$i] = $E1[$n-1][$$edges[$ei]{'S'}] + $sub_cost[$n][$i]; 
	  $be[$n][$i] = $ei; $bl[$n][$i] = $n-1;
	}

	# Second, look back at same ref level: is insertion_cost best so far?
	if(($E1[$n][$$edges[$ei]{'S'}] < $E1[$n][$i] - $NODE_INSERTION_COST)) {
	  $E1[$n][$i] = $E1[$n][$$edges[$ei]{'S'}] + $NODE_INSERTION_COST; 
	  $be[$n][$i] = $ei; $bl[$n][$i] = $n;
	}
      }
    }
  }
    
  #######
  # Backward cost computation
  # Initialize: at level $#times, 
  #   cost to each node determined by number of insertions in path ahead
  my(@E2) = ();   $E2[$#times][$#$nodes] = 0;
  my(@fl) = ();   $fl[$#times][$#$nodes] = $#times;
  my(@fe) = ();   $fe[$#times][$#$nodes] = -1;
  for ($i=$#$nodes; $i >= 0; $i--) {
    $E2[$#times][$i] = 100*$NODE_INSERTION_COST;
    foreach $ei (@{$$nodes[$i][1]}) {
      if($E2[$#times][$$edges[$ei]{'E'}] < 
	 $E2[$#times][$i] - $NODE_INSERTION_COST) {
	$E2[$#times][$i] = $E2[$#times][$$edges[$ei]{'E'}] + 
	  $NODE_INSERTION_COST; 
	$fe[$#times][$i] = $ei; 
	$fl[$#times][$i] = $#times;
      }
    }
  }

  # Now compute minimum-cost path at each lower level
  for ($n=$#times-1; $n>=0; $n--) {
    for ($i=$#$nodes; $i >= 0; $i-- ) {
      # $E2[$n][$i] is no larger than $E2[$n+1][$i] + one deletion
      $E2[$n][$i] = $E2[$n+1][$i] + $NODE_DELETION_COST;
      $fl[$n][$i] = $n+1; $fe[$n][$i] = -1;
      
      # Search through all edges going forward from this node
      foreach $ei (@{$$nodes[$i][1]}) {
	# First, look up and forward: is sub_cost less than current best?
	if($E2[$n+1][$$edges[$ei]{'E'}] < 
	   $E2[$n][$i] - $sub_cost[$n+1][$$edges[$ei]{'E'}]) {
	  $E2[$n][$i] = $E2[$n+1][$$edges[$ei]{'E'}] + 
	    $sub_cost[$n+1][$$edges[$ei]{'E'}]; 
	  $fl[$n][$i] = $n+1; $fe[$n][$i] = $ei;
	}
	
	# Second, look forw at same ref level: is insertion_cost best so far?
	if(($E2[$n][$$edges[$ei]{'E'}] < $E2[$n][$i] - $NODE_INSERTION_COST)) {
	  $E2[$n][$i] = $E2[$n][$$edges[$ei]{'E'}] + $NODE_INSERTION_COST; 
	  $fl[$n][$i] = $n; $fe[$n][$i] = $ei;
	}
      }
    }
  }

  # Now find best alignment for each node
  my(@align) = ();
  foreach $i (0..$#$nodes) {
    # Initialize: try best alignment is between $$time[0] and $$time[1]
    my($best_cost) = $E1[0][$i] + $E2[0][$i];
    $align[$i]= ($bl[0][$i] < 0) ? $$mapnodes[0] : -1;
    # Now see if any higher level gives lower cost
    foreach $n (1..$#times) {
      if($E1[$n][$i]+$E2[$n][$i] < $best_cost) {
	$best_cost = $E1[$n][$i] + $E2[$n][$i];
	$align[$i]= ($bl[$n][$i] < $n) ? $$mapnodes[$n] : -1;
      }
    }
  }

  # For all nodes that were aligned "-1": 
  #    Add an insertion node to the MAP path,
  #    or align with an existing insertion node if one is close enough
  my($NMAP) = @times;
  foreach $i (0..$#$nodes) {  
    if($align[$i] < 0) {
      # Search the already-existing insertion nodes, if any,
      #   to find the one that is closest to node $i
      my($best_cost) = $NODE_INSERTION_COST;
      my($best_level) = -1;
      foreach $n ($NMAP..$#times) {
	my($cost) = abs($$nodes[$i][2] - $times[$n]);
	if($cost < $best_cost) { $best_cost=$cost; $best_level=$n; }
      }
      # If best cost is below threshold, align this node with $n
      if($best_cost < $INSERTION_ALIGNMENT_THRESHOLD) {
	$align[$i] = $$mapnodes[$best_level];
      }
      # Otherwise, create a new insertion node
      else {
	push(@times, $$nodes[$i][2]);
	push(@$mapnodes, $i);
	$align[$i] = $i;
      }
    }
  }

  return(\@align, $mapnodes);
}



##################################################
# Main program

if (@ARGV==0) { print $USAGE; exit; }

# Read in the script file
$arg=shift;
open(F,$arg) || die "Unable to read from script file $arg: $!\n";
$script = ''; while (<F>) { $script .= $_; }
close(F);

# Main loop: process each pair of lattice files in the script
while($script =~ s/^\s*(\S+)\s+(\S+)//) {
  my($inlat) = $1; my($outlat) = $2;

  # Read the input lattice
  unless(open(LATFILE, $inlat)) {
    print STDERR "Unable to read from lattice $inlat: $!\n";
    next;
  }
  my(@graph) = read_HTK_lattice(\*LATFILE);
  close(LATFILE);

  # First, find all edges that are tagged with "n=1",
  #   store their node numbers as keys in %mapnodes,
  #   store the node times as values in %mapnodes,
  #   this strategy avoids duplicate entry of the same node number
  my(%mapnodekeys) = ();
  foreach $ei (0..$#{$graph[1]}) {
    if ($graph[1][$ei]{'n'} eq "1") {
      $mapnodekeys{$graph[1][$ei]{'S'}} = 1;
      $mapnodekeys{$graph[1][$ei]{'E'}} = 1;
    }
  }

  # Compute alignment between the MAP nodes and all other nodes
  #   Make sure it's sorted
  my($mapnodes) = [ sort {$a <=> $b} keys %mapnodekeys ];
  ( $align, $mapnodes ) = dp_align($graph[0], $graph[1], $mapnodes);
  $mapnodes = [ sort {$a <=> $b} @$mapnodes ];

  # Create a table: MAP node is the key, value is the list
  #   of original nodes aligned with the MAP node
  my(%map2nodes) = (); my($m); my($i);
  foreach $m (@$mapnodes) { $map2nodes{$m} = []; }
  foreach $i (0..$#$align) { push(@{$map2nodes{$$align[$i]}}, $i); }
  
  # Create another table, giving compacted index of mapnode
  #  as a function of its original index
  foreach $m (0..$#$mapnodes) { $old2new{$$mapnodes[$m]}=$m; }

  # Create a new graph, with same header as original
  my(@newgraph) = ( [], [], $graph[2] );
  # Create brand new nodes, but with times matching @$mapnodes
  foreach $m (0..$#$mapnodes) 
    { $newgraph[0][$m] = [ [], [], $graph[0][$$mapnodes[$m]][2] ]; }

  # Now create all of the edges
  foreach $m (0..$#$mapnodes) {
    # Copy all edges that start at any of the nodes aligned with $$mapnodes[$m]
    foreach $i (sort {$a <=> $b} @{$map2nodes{$$mapnodes[$m]}}) {
      # Count through the forward edgelist of this node
      my($ei);
      foreach $ei (@{$graph[0][$i][1]}) {
	my($newstart) = $old2new{$$align[$graph[1][$ei]{'S'}]};
	my($newend) = $old2new{$$align[$graph[1][$ei]{'E'}]};

	# New start and end nodes of this edge are the new indices
	#  of the MAP nodes to which 
	#   the original start and end nodes were aligned
	$graph[1][$ei]{'S'} = $newstart;
	$graph[1][$ei]{'E'} = $newend;

	# If any previously viewed edges start at the same node,
	#   end at the same end node, and carry same word, 
	#   merge probabilities of the two edges,
	#    and mark edge for deletion by setting $e{'E'}=0
	my($fi);
	foreach $fi (@{$newgraph[0][$m][1]}) {
	  if($newgraph[1][$fi]{'E'} eq $graph[1][$ei]{'E'} &&
	     $newgraph[1][$fi]{'W'} eq $graph[1][$ei]{'W'}) {
	    $newgraph[1][$fi]{'a'} +=
	      log(1+exp($graph[1][$ei]{'a'}-$newgraph[1][$fi]{'a'}) );
	    $newgraph[1][$fi]{'l'} +=
	      log(1+exp($graph[1][$ei]{'l'}-$newgraph[1][$fi]{'l'}));
	    if($graph[1][$ei]{'n'} eq '1') { $newgraph[1][$fi]{'n'} = "1"; }
	    $graph[1][$ei]{'E'} = 0;
	  }
	}
	
	# If the edge duration is positive, push it onto the graph,
	#  and also to the forward and backward edgelists of its endpoints
	if($newgraph[0][$graph[1][$ei]{'S'}][2] < 
	   $newgraph[0][$graph[1][$ei]{'E'}][2]) {
	  push(@{$newgraph[1]}, $graph[1][$ei]);
	  push(@{$newgraph[0][$graph[1][$ei]{'S'}][1]}, $#{$newgraph[1]});
	  push(@{$newgraph[0][$graph[1][$ei]{'E'}][0]}, $#{$newgraph[1]});
	}
      }
    }
  }

  # Write the modified lattice
  unless(open(LAT_OUT, ">$outlat")) {
    print STDERR "Unable to write lattice out to $outlat: $!\n";
    next;
  }
  write_HTK_lattice(@newgraph, \*LAT_OUT);
  close(LAT_OUT);

  # Write summary to STDERR
  my($nn)=0+@{$newgraph[0]};
  my($ne)=0+@{$newgraph[1]};
  print STDERR "$inlat -> $outlat, $nn nodes, $ne edges\n";
}

