#!/usr/bin/perl
# lattice_cheat.pl: Mark Hasegawa-Johnson, October 2003
# Illinois Speech and Language Engineering (ISLE group)
#
# Usage text (printed if lattice_cheat.pl is called with no arguments):
$USAGE=<<ENDUSAGE;

lattice_cheat.pl: Rescore lattice using known subword transcriptions
  plus a dictionary.

Usage
  lattice_cheat.pl PE MLF DICT DEBUG < script

Example
  lattice_cheat.pl 0.2 ref_lms.mlf pronlex_lw.dict < rescore_lw.scp
  lattice_cheat.pl 0.2 ref_phones.mlf pronlex.dict < rescore_pw.scp

lattice_cheat.pl augments the acoustic model score, on each arc in a
lattice, by a score computed from the given transcriptions in MLF
combined with the pronunciation model in DICT.  The acoustic model
score is incremented by 100*log(sum_{i=1}^N p(pron_i|word)
PE^L(pron_i)), where p(pron_i|word) is the probability of the ith
pronunciation, PE is the error probability given on the command line,
L(pron_i) is the Levenshtein distance between pronunciation-#i and the
given subword transcription, and multiplication by 100 is a heuristic
attempt to make the probability scales comparable.

The script file should contain input & output filenames in pairs,
i.e., "input1.lat output1.lat input2.lat output2.lat ..."

If the fourth argument, DEBUG, is specified and nonzero, then detailed
trace information is printed for each lattice edge scored.

ENDUSAGE

####################################################################

####################################################################
# GLOBAL VARIABLES

# Conversion rates for various file formats
$MLFPERIOD = 0.0000001;

# Math functions
sub abs { my($x)=shift; ($x>0) ? $x : -$x ; }
sub round { $_[0] =~ /([^\.]*)(.*)/; ($2 >= 0.5) ? $1+1 : $1; }

 
###################################################
# Subroutine to read MLF entry
#
# Syntax: $data = read_MLF_entry("filename", \*INPUTFILE);
#
# Returned value $data is an arrayref of arrayrefs
# $$tran[0][0] is first seg start time, $$tran[0][1] is end time,
# @$$tran[0][2..$#{$$tran[0]}] is the sequence of labels at first time.
#
sub read_MLF_entry {
  my($filename) = $_[0];
  my($fh) = $_[1];
  my(@tran) = ();
  my(@labels) = ();
  
  # Read all continuous lines that begin with a char other than period
  for($l=<$fh>; $l =~ /^[^\.]/; $l=<$fh>) {
    # Option 1: line is in "time time labels" format
    if ($l=~/(\d+)\s+(\d+)\s+(.*)/ ) {
      @labels = split(/\s+/,$3);
      push(@tran, [ $1 * $MLFPERIOD, $2 * $MLFPERIOD, @labels ] );
    }
    # Option 2: line contains any non-whitespace labels, but no times
    elsif ($l =~ /\S/) {
      @labels = split(/\s+/,$l);
      push(@tran, [ 0, 0, @labels ] );
    }
  }
  return(\@tran);
}

###################################################
# Subroutine to read a noway-format (or HTK-format) dictionary
#
# Syntax: $dict = read_dict(\*FILEHANDLE)
#
# $dict = hashref of arrayrefs of arrayrefs
#   @{$$dict{'word'}[0]} is the first pronunciation of "word", e.g., 
#   $$dict{'word'}[0] = [ 0.35 w er d ];
#   where the first entry is the probability of this pronunciation.
#
# If entries have no attached probabilities (e.g., HTK dictionaries),
# all probabilities in dict are set to 1.0.
#
sub read_dict {
  my($fh) = $_[0];
  my(%dict) = ();
  my($word) = '';

  # Read every line, ignoring blank lines
  while(($line=<$fh>) !~ /^$/) {
    chomp($line); chomp($line);
    my(@pron) = split(/\s+/,$line);

    # Case 1: first field is word(prob)
    if($pron[0] =~ /(\S+)\(([\d\.]+)\)/) {
      $word = $1;
      $pron[0] = $2;
    }
    # Case 2: first field is just word; set prob to 1.0
    else {
      $word = $pron[0];
      $pron[0] = 1;
    }
    if(exists($dict{$word})) { push(@{$dict{$word}}, \@pron); }
    else { $dict{$word} = [ \@pron ]; }
  }
  return(\%dict);
}

###################################################
# Subroutine to read an HTK lattice file
#
# Syntax: ($nodes, $edges, $header) = read_HTK_lattice(\*FILEHANDLE)
#
# $nodes is an arrayref of nodes
#    each node is an arrayref containing the following data:
#    [ back, forw, time ]
#    where forw is an arrayref of indices of forward edges,
#    back is an arrayref of indices of backward edges.
#    time is the time-alignment of this node
# $edges is an arrayref of edges
#    each edge is an arrayref as follows:
#    [ sn, en, word, ap, lp ]
#    sn is the index of the start node
#    en is the index of the end node
#    word is the word on this node
#    ap is the acoustic probability
#    lp is the language model probability
# $header is a hashref containing the header information
#
sub read_HTK_lattice {
  my($fh) = $_[0];
  my(@nodes) = ();
  my(@edges) = ();
  my(%header) = ();
  my($line);

  # Read lines of the file
  while($line = <$fh>) {
    # Node number given by I=
    if($line =~ /I=(\d+)/) {
      my($nodenum) = $1;
      unless(exists($nodes[$nodenum])) { $nodes[$nodenum] = [ [],[],0 ]; }
      if($line =~ /t=([\.\d]+)/) { $nodes[$nodenum][2] = $1; }
    }
    # Edge number given by J=
    elsif($line =~ /J=(\d+)/) {
      my($edgenum) = $1;
      my(@edge) = (0,0,'',0,0);
      if($line =~ /S=(\d+)/) { 
	$edge[0] = $1;
	# Tell the start node that I am one of its forward edges
	push(@{$nodes[$1][1]}, $edgenum);
      }
      if($line =~ /E=(\d+)/) { 
	$edge[1] = $1; 
	# Tell the end node that I am one of its backward edges
	push(@{$nodes[$1][0]}, $edgenum);
      }
      # Get the word label
      if($line =~ /W=(\S+)/) { $edge[2] = $1; }
      # Get the acoustic score
      if($line =~ /a=([\-\.\d]+)/) { $edge[3] = $1; }
      # Get the language model score
      if($line =~ /l=([\-\.\d]+)/) { $edge[4] = $1; }
      $edges[$edgenum] = \@edge;
    }
    # Otherwise, try to read header information
    else {
      while ($line =~ s/(\S+)=(\S+)//) { $header{$1} = $2; }
    }    
  }

  return(\@nodes, \@edges, \%header);
}

###################################################
# Subroutine to write an HTK lattice file
#
# Syntax: write_HTK_lattice($nodes, $edges, $header, \*FILEHANDLE)
#
# $nodes is an arrayref of nodes
# $edges is an arrayref of edges
# $header is a hashref containing HTK lattice header information
# FILEHANDLE is the opened output file
#
# $N is the number of nodes actually written to the file
# $L is the number of edges actually written to the file
#
sub write_HTK_lattice {
  my($nodes) = $_[0];
  my($edges) = $_[1];
  my($header) = $_[2];
  my($fh) = $_[3];
  my($N) = 0;
  my($L) = 0;

  # Print header information
  print $fh "VERSION=1.0\n";
  print $fh "UTTERANCE=",$$header{'UTTERANCE'},"\n";
  print $fh "lmname=",$$header{'lmname'},"\n";
  print $fh "lmscale=",$$header{'lmscale'}," ";
  print $fh "wdpenalty=",$$header{'wdpenalty'},"\n";
  print $fh "vocab=",$$header{'vocab'},"\n";
  print $fh sprintf("N=%-5d L=%-7d\n",0+@{$nodes},0+@{$edges});

  # Print nodes
  foreach $N (0..$#{$nodes}) {
    print $fh sprintf("I=%-5d ",$N);
    if (exists($$nodes[$N][2])) 
      { print $fh sprintf("t=%4.2f",$$nodes[$N][2]); }
    print $fh "\n";
  }
  # Print edges
  foreach $L (0..$#{$edges}) {
    print $fh sprintf("J=%-7d S=%-5d E=%-5d W=%-20s a=%7.7f l=%7.7f\n",
		      $L, @{$$edges[$L]});
  }
}

###################################################
# Extract word string from an MLF-style transcription
#
# $wrds = trans2wrds($trans)
#
# $wrds: arrayref of wrds
# $trans: arrayref of transcriptions
#
sub trans2wrds {
  my(@trans) = @{$_[0]};
  my(@wrds) = ();
  foreach $t (@trans) { push(@wrds, $$t[2]); }
  return(\@wrds);
}

###################################################
# Find Levenshtein statistics between two symbol strings
#
# $stats = levenshtein_skipN( $ref, $test, $skippable )
#
# $stats = [ N, S, D, I, E=S+D+I ]
# $ref = reference string
# $test = test string
# $skippable = number of words that can be skipped without loss
#      at start or end of $test
#
# N counts the number of non-null wrds in $ref.
# S, D, I = number of subs, dels, ins, and total errors
#   in the best alignment of $ref with $test.
#   Insertions and deletions do not count null wrds ('').
#   E=S+D+I.
#
sub levenshtein_skipN {
  my($ref) = $_[0];
  my($test) = $_[1];
  my($skippable) = $_[2];
  # Stats vector keeps track of output stats, and also, the backpointers
  my(@stats) = ();
  my($m); 
  my($n);
  my($M)=$#{$ref}+1;
  my($N)=$#{$test}+1;
  my($numref)=0; 
  my($numtest)=0; 

  # If zero wrds observed in test list, (SDIE) = (0,D,0,D,N-1,M-1)
  foreach $m (0..$M) { 
    if( $m>0 && $$ref[$m-1] ne '') { $numref++ };
    $stats[$m][0] = [0,$numref,0,$numref,$m-1,0]; 
  }

  # If zero wrds observed in ref list, (SDIE) = (0,I,I)
  foreach $n (0..$N) { 
    if( $n>0 && $$test[$n-1] ne '') { $numtest++ };
    # Allow skipping of the first $skippable words in $ref
    $insertions = ($numtest>$skippable)?($numtest-$skippable):0;
    $stats[0][$n] = [0,0,$insertions,$insertions,0,$n-1]; 
  }

  # m is index into the ref list, n is index into the test list    
  foreach $m (1..$M) { foreach $n (1..$N) {

    # If wrds are equal, there is no cost; 
    #   Otherwise, we suffer at least one substitution
    @{$stats[$m][$n]} = ( @{$stats[$m-1][$n-1]}[0..3], $m-1, $n-1 );
    unless($$ref[$m-1] eq $$test[$n-1]) 
      { $stats[$m][$n][0]++; $stats[$m][$n][3]++; }
    
    # If $ref[$m-1] eq '', we can skip it without cost
    if ($$ref[$m-1] eq '' && $stats[$m-1][$n][3] < $stats[$m][$n][3]) 
      { @{$stats[$m][$n]} = ( @{$stats[$m-1][$n]}[0..3], $m-1, $n ); }
    # Otherwise, if deletion results in minimum cost, use it
    elsif ($stats[$m-1][$n][3] < $stats[$m][$n][3] - 1) {
      @{$stats[$m][$n]} = ( @{$stats[$m-1][$n]}[0..3], $m-1, $n );
      $stats[$m][$n][1]++; $stats[$m][$n][3]++;
    }
    
    # If $test[$n-1] eq '', we can skip it without cost
    if ($$test[$n-1] eq '' && $stats[$m][$n-1][3] < $stats[$m][$n][3]) 
      { @{$stats[$m][$n]} = ( @{$stats[$m][$n-1]}[0..3], $m, $n-1 ); }
    # Otherwise, if insertion results in lower cost, use it
    elsif ($stats[$m][$n-1][3] < $stats[$m][$n][3] - 1) {
      @{$stats[$m][$n]} = ( @{$stats[$m][$n-1]}[0..3], $m, $n-1 );
      $stats[$m][$n][2]++; $stats[$m][$n][3]++;
    }
  }}

  # Allow skipping of the last $skippable words in $test
  my($lastword) = $N;
  foreach $i (1..$skippable) {
    if($stats[$M][$N-$i][3] < $stats[$M][$N][3]) { $lastword = $N-$i; }
  }

  # Debug: backtrace the list of word alignments
  if($DEBUGLEVEL) {
    my($backm) = $M;
    my($backn) = $lastword;
    my(@backtrace) = ();
    while($backm > 0 || $backn > 0) {
      unshift(@backtrace,"");
      if($backm != $stats[$backm][$backn][4] && $backm>0) 
	{ $backtrace[0] = $$ref[$backm-1]; }
      else { $backtrace[0] = "<I>"; }
      if($backn != $stats[$backm][$backn][5] && $backn>0) 
	{ $backtrace[0] .= "," . $$test[$backn-1]; }
      else { $backtrace[0] .= ",<D>"; }
      $backm = $stats[$backm][$backn][4];
      $backn = $stats[$backm][$backn][5];
    }
    # Debug: print out the backtrace
    print sprintf("(N%d,S%d,D%d,I%d,E%d,LW%d-%d):",
		  $numref,$stats[$M][$lastword][0],$stats[$M][$lastword][1],
		  $stats[$M][$lastword][2],$stats[$M][$lastword][3],
		  $N,$N-$lastword);
    foreach $i (@backtrace) { print " ($i)"; }  print "\n";
  }

  # Return (N, S, D, I, E) of the last word
  return( $numref, @{$stats[$M][$lastword]}[0..3] );
}

###################################################
# normalize_word:
# Remove distinctions not useful for scoring a lattice
# from a word in input transcription or on a lattice edge.
#
# This function should be same as normalize_words.pl,
# except that noise, and laughter are replaced with a null string "".
#
# $outword = normalize_word($inword)
#
sub normalize_word {
  # Set the word to upper case
  my($l) = uc($_[0]);
    
  # Any compound including the word "laugh" means "laugh"
  $l =~ s/.*LAUGH.*//;
  
  # These words seem to be used for various filled pauses
  #   Replace all with UH or UM
  $l =~ s/.*(EPENTHETIC|FILLEDSIL|HESITATION|GLOTTAL).*//;
  $l =~ s/.*(TRUNCATEDP|FALSESTART).*//;
  $l =~ s/.*(NASAL|MMM).*//;
  $l =~ s/\b(UH|UM|UMM)\b//;

  # Eliminate various background noise notations
  $l =~ s/\bSIL\b//;
  
  # These words mean silence even if extended, e.g. SILENCEDUETOSPLICE
  $l =~ s/.*(H\#|CROFF|UNKNOWN|CLICK|SILENCE|BREATH).*//;
  $l =~ s/.*(CROSSTALK|STATIC|NOISE|SMACK|BUZZ|HUMM).*//;
  
  # Replace periods by apostrophes
  $l =~ s/\./\'/g;
  # Eliminate second in any series of apostrophes
  $l =~ s/\'\'/\'/g;    
  
  # Eliminate commentary between brackets, braces, or parentheses,
  #  after underscore, or after ? or ! or #
  # This should come last, so that # doesn't keep H# from being used
  $l =~ s/\[.*\]//;
  $l =~ s/\{.*\}//;
  $l =~ s/\(.*\)//;
  $l =~ s/(_|\!|\?|\#).*//;
  
  # Eliminate quotation marks
  $l =~ s/\"//g;
  
  # Return outword
  return($l);
}

################################################################
################################################################
# Main program

if (@ARGV==0) { print STDERR $USAGE; exit; }

# FIRST ARGUMENT:
#    Get the  error probability
$LOGERRORPROB = log($ARGV[0]);

# SECOND ARGUMENT:
#    Read in all lines from the reference MLF
%IMLF  = ();
open(MLF_FILE, $ARGV[1]) || die "Unable to open MLF $ARGV[1]: $!";
while (<MLF_FILE>) {
  # If this line starts a transcription, read it
  if (/\".*\/([^\/\.]+)[^\/]*\"/) {
    my($filename) = $1;
    $IMLF{$filename} = read_MLF_entry($filename,\*MLF_FILE);
  }
}
close(MLF_FILE);

# THIRD ARGUMENT:
#     Read in the dictionary
open(DICT_FILE, $ARGV[2]) || die "Unable to open DICT $ARGV[2]: $!";
$DICT = read_dict(\*DICT_FILE);
close(DICT_FILE);

# FOURTH ARGUMENT:
#      Set the DEBUG level
$DEBUGLEVEL = $ARGV[3];

# STANDARD INPUT:
#       Read in the script file
$script = ''; while (<STDIN>) { $script .= $_; }

#######################
# Main loop: process each pair of files in the script
while($script =~ s/^\s*(\S+)\s+(\S+)//) {
  $inlat = $1; $outlat = $2;

  # If input lattice can't be read, skip to next pair of filenames
  next unless open(LATFILE, $inlat);

  # Read the entire graph from the lattice file
  my(@graph) = read_HTK_lattice(\*LATFILE);
  close(LATFILE);
  unless(exists($graph[0][0][2])) {
    print STDERR "Node times not specified in lattice $inlat\n";
    next;
  }
  
  # Get MLFname from $header{'UTTERANCE'}; strip extensions & dirs
  $MLFname = $graph[2]{'UTTERANCE'}; 
  $MLFname =~ s/(\S+\/)|(\.\S+)//g;

  # Look up corresponding transcription in IMLF
  next unless exists($IMLF{$MLFname});
  $trans = $IMLF{$MLFname};
  $words = trans2wrds($trans);

  # Set up records to keep track of initial and final total scores
  $starting_total_score = 0;
  $ending_total_score = 0;

  # Write a summary to STDERR
  print STDERR "$inlat -> $outlat\n";

  # Evaluate every edge in the lattice
  foreach $ei (0..$#{$graph[1]}) {
    # Get the word from this edge
    $edgeword = normalize_word($graph[1][$ei][2]);

    # Find times of the nodes
    $t0 = $graph[0][$graph[1][$ei][0]][2];
    $t1 = $graph[0][$graph[1][$ei][1]][2];
    
    # Extract transcription between $t0 and $t1,
    $n0 = 0; while($n0+1 <= $#$trans && $$trans[$n0][1] < $t0) { $n0++; }
    $n1 = $#$trans; while($n1-1 >= 0 && $$trans[$n1][0] > $t1) { $n1--; }
    #  plus one label outside of [$t0,$t1] on either side
    #$n0 = 0; while($n0+1 <= $#$trans && $$trans[$n0+1][1] < $t0) { $n0++; }
    #$n1 = $#$trans; while($n1-1 >= 0 && $$trans[$n1-1][0] > $t1) { $n1--; }
    @subwords = @$words[$n0..$n1];

    # Add up the scores of all alternate pronunciations,
    #  penalized by the number of errors;
    #  allow skip of 2 words at start or end of transcription
    $cum_pdpe = 0;
    $cum_pd = 0;
    $DEBUGLEVEL && print "$ei: $edgeword\n";

    # If word has been normalized away by normalize_word,
    #   then anything in the transcription should be counted as error;
    #   thus we count real errors only if $edgeword ne ''
    if($edgeword && exists($$DICT{$edgeword})) {
      foreach $pron (@{$$DICT{$edgeword}}) { 
	# test_pron contains the pronunciation, $pd is the probability
	@test_pron = ( @{$pron} );
	$pd = shift(@test_pron);
	$cum_pd += $pd;
	
	# Compute Levenshtein distance from subwords to test_pron
	@stats = levenshtein_skipN(\@test_pron, \@subwords, 0);

	# Total prob = dictionary prob * (ERRORPROB)^(NUM_ERRORS)
	$pdpe = $pd * exp($stats[4] * $LOGERRORPROB);
	$cum_pdpe += $pdpe;
	$DEBUGLEVEL && 
	  print sprintf("P(d)P(e)=%g * PE^%d=%g, sum(pd)=%g, sum(pdpe)=%g\n",
			$pd, $stats[4], $pdpe, $cum_pd, $cum_pdpe);
      }
    }
    # Record the edge score prior to augmentation
    $DEBUGLEVEL && print "$graph[1][$ei][3] ->";
    $starting_total_score += $graph[1][$ei][3];

    # If score was computed, add it to edge
    if($cum_pd > 0 && $cum_pdpe > 0)
      { $graph[1][$ei][3] += 100 * log($cum_pdpe/$cum_pd); }
    # If the arc contained an unkown word, 
    #   penalize it severely, as a heuristic to avoid 
    #   uncontrollable errors from unkown words
    else { 
      # $graph[1][$ei][3] += 100*($#$subwords+1)*$LOGERRORPROB; 
      $graph[1][$ei][3] -= 100000;
    }

    # Record the edge score after augmentation
    $DEBUGLEVEL && print " $graph[1][$ei][3]\n\n";
    $ending_total_score += $graph[1][$ei][3];
  }

  # Write the modified graph
  unless(open(LAT_OUT, ">$outlat")) {
    print STDERR "Unable to write lattice out to $outlat: $!";
    next;
  }
  write_HTK_lattice(@graph, \*LAT_OUT);
  close(LAT_OUT);

  # Write a summary to STDERR
  print STDERR sprintf("$inlat: %g -> %g (%d,%d)\n",
		       $starting_total_score,$ending_total_score,
		       $#{$graph[0]}+1, $#{$graph[1]}+1);
}


