#!/usr/bin/perl
# rnc2txgd.pl: Mark Hasegawa-Johnson, 5/17/2005
# Illinois Speech and Language Engineering (ISLE group)
#
# Usage (printed if called with no arguments):
#
$USAGE=<<ENDUSAGE;

RNC2TXGD: Convert ESPS transcription to Praat TextGrid
 
Usage:
  rnc2txgd.pl file1.esps:usage1 file2.esps:usage2 file3.esps:usage3 outfile.TextGrid
Example:
  rnc2txgd.pl f1ajrlp1.wrd:s f1ajrlp1.brk:p f1ajrlp1.ton:p f1ajrlp1.TextGrid

This script reads in a series of files in ESPS file format.  Each is
converted into a TextGrid transcription tier, and the tier is written
out.

"usage" determines whether the corresponding TextGrid tier should be a
segment tier (s) or a point tier (p).  If usage is not specified, the
default is segment.

ENDUSAGE

####################################################################
# GLOBAL VARIABLES

# Conversion rate for SPHERE file format
$ESPSPERIOD = 1;

# Math functions
sub abs { my($x)=shift; ($x>0) ? $x : -$x ; }


###################################################
# Subroutine to read RNC-format transcription file (TON, BRK, or WRD)
#
# Syntax: $data = read_RNC_file("filename");
#
# $data is an arrayref of arrayrefs.
# $$data[$i][0] is the start time of the $ith label
# $$data[$i][1] is the end time of the $ith label
# @$$data[$i][2..$#{$$tran[0]}] is the sequence of labels at time $i.
#
sub read_RNC_file {
  # Initialize @data to null
  my(@data) = ();

  # Open the INPUTFILE or die with an error message
  open(INPUTFILE,$_[0]) || die "Unable to open input file $_[0]: $!";

  # Read in all lines from the INPUTFILE
  foreach $_ (<INPUTFILE>) {
    # Ignore lines not in the format of number, number, labels
    if ( /^\s*([\d\-\.]+)\s+([\d\-\.]+)\s+(.*)/ ) {
      my(@labels) = split(/\s+/,$3);

      # Push a reference to @record onto the @data array
      push(@data, [ 0, $1 * $ESPSPERIOD, @labels ] );
    }
  }
  close(INPUTFILE);

  # Set start time of each segment to end time of preceding
  foreach $n (1..$#data) 
    { ${$data[$n]}[0] = ${$data[$n-1]}[1]; }

  return(\@data);
}

###################################################
# Subroutine to write a TextGrid header
#
# Syntax: write_textgrid_header($fh, $firsttime, $lastttime, $maxdepth)
#
sub write_textgrid_header {
  my($fh) = $_[0];
  my($firstttime) = $_[1];
  my($lasttime) = $_[2];
  my($maxdepth) = $_[3];

  # Print the TEXTGRID header
  print $fh 'File type = "ooTextFile"',"\n";
  print $fh 'Object class = "TextGrid"',"\n\n";
  print $fh "xmin = $firstttime\n";
  print $fh "xmax = $lasttime\n";
  print $fh 'tiers? <exists>',"\n";
  print $fh "size = $maxdepth\n";
  print $fh 'item []: ',"\n";
}

###################################################
# Print a TextGrid tier
#
# Syntax: sprint_textgrid_tier(\@data, $number, $starttime, $endtime)
#
sub sprint_textgrid_tier {
  # Use a private version of phndata array,
  # so we can insert elements w/o global consequences
  my(@phndata) = @{$_[0]};  
  my($number) = $_[1];
  my($starttime) = $_[2];
  my($endtime) = $_[3];
  my($usage) = $_[4];

  # If tier is null, create a one-segment tier
  unless(@phndata) { @phndata = ([$starttime, $endtime, ""]); }
    
  # First: insert segments so that segments are collectively exhaustive
  # This must be done first, so that array size is known in advance
  if($phndata[$#phndata][1] < $endtime) 
    { push(@phndata,[$phndata[$#phndata][1],$endtime,""]); }
  for($nt=$#phndata; $nt>0; $nt--) {
    if($phndata[$nt-1][1] < $phndata[$nt][0]) 
      { splice(@phndata,$nt,0,[$phndata[$nt-1][1],$phndata[$nt][0],"" ]) }
  }
  if($phndata[0][0] > $starttime) 
    { unshift(@phndata,[$starttime,$phndata[0][0],""]); }

  # Start with "item [$number]"
  my($output) = "    item [$number]:\n";

  # If usage is "p", write a point tier/Text Tier
  if($usage eq "p") {
      $output .= "        class = \"TextTier\"\n"; 
      $output .= "        name = \"tier${number}\"\n"; 
      $output .= "        xmin = $starttime\n";
      $output .= "        xmax = $endtime\n";
      $output .= sprintf("        points: size = %d\n", $#phndata + 1);

      # Print out the converted data 
      foreach $nt (0..$#phndata) {
	  $output .= sprintf("        points [%d]:\n",$nt+1);
	  $output .= "            time = $phndata[$nt][1]\n";
	  $output .= "            mark = \"$phndata[$nt][2]\"\n";
      }
  }
  # Otherwise: if usage is not "p", write an interval tier
  else {
      $output .= "        class = \"IntervalTier\"\n"; 
      $output .= "        name = \"tier${number}\"\n"; 
      $output .= "        xmin = $starttime\n";
      $output .= "        xmax = $endtime\n";
      $output .= sprintf("        intervals: size = %d\n", $#phndata + 1);

      # Print out the converted data 
      foreach $nt (0..$#phndata) {
	  $output .= sprintf("        intervals [%d]:\n",$nt+1);
	  $output .= "            xmin = $phndata[$nt][0]\n";
	  $output .= "            xmax = $phndata[$nt][1]\n";
	  $output .= "            text = \"$phndata[$nt][2]\"\n";
      }
  }

  return($output);
}


###########################################
# Main Program

if (@ARGV < 2) { print $USAGE; exit; }

# Read in the input files
#
$firsttime=100000;
$lasttime=-100000;
foreach $i (0..($#ARGV-1)) {
    my($filename);
    if ($ARGV[$i] =~ /(.*):(.*)/) {
	$filename=$1;
	$usage[$i]=$2;
    }
    else {
	$filename=$ARGV[$i];
	$usage[$i]='s';
    }
    $data[$i] = read_RNC_file($filename);
    # Check to see if this transcription starts earlier or ends later
    # than previously read transcriptions
    if($data[$i][0][0] < $firsttime) {
	$firsttime = $data[$i][0][0];
    }
    if($data[$i][$#{$data[$i]}][1] > $lasttime) {
	$lasttime = $data[$i][$#{$data[$i]}][1];
    }
}

# Open the TEXTGRID file and write its header
open(TEXTGRID, ">$ARGV[$#ARGV]") || die "Unable to write to $ARGV[$#ARGV]: $!";
write_textgrid_header(\*TEXTGRID, $firsttime, $lasttime, $#data + 1);

# Write tiers to textgrid file
foreach $i (0..$#data) {
    print TEXTGRID sprint_textgrid_tier($data[$i], $i, $firsttime, $lasttime, $usage[$i]);
}

close(TEXTGRID);


