#!/usr/bin/perl
# syllabify_edges.pl: Mark Hasegawa-Johnson, March 2004
# Illinois Speech and Language Engineering
#
# Usage text (printed if divide_edges.pl is called with no arguments):
$USAGE=<<ENDUSAGE;

syllabify_edges.pl: Divide pronunciation model of an edge into
syllables.  At syllable boundaries in the pronunciation model, replace
the colon separator (:) by a semicolon-colon sequence (;:).

Usage:
 syllabify_edges.pl script.scp

The script file should contain input & output filenames in pairs,
i.e., "input1.lat output1.lat input2.lat output2.lat ...".  Each input
lattice file should contain a word lattice for exactly one utterance.
Every edge in the word lattice should have a p= tag, specifying the
pronunciation of that edge.  

The pronunciation model should be a sequence of landmark specifiers.

This program divides the pronunciation into syllables.  Each +syllabic
landmark is the core of a syllable.  

We assume that the pronunciation model doesn't specify
syllabification, and therefore syllabification of consonants must be
determined by rule.  The following rules are used:

1. The first consonant closure goes into coda of the first syllable.
The closure is coda, even if release of the same consonant is in the
onset of the following syllable.  This is an explicit representation
of ambisyllabicity --- the appropriateness of this representation
depends on the application!

2. As many as possible of the remaining landmarks are allocated to the
onset.  The first consonant release or glide (-syllabic landmark) to
the left of the vowel is always part of the onset.  Any obstruent
preceding an /r/ may join it in the onset.  A non-blade obstruent
preceding an /l/ may join it in the onset.  A non-lips obstruent
preceding a /w/ may join it in the onset.  An /s/ preceding any
voiceless stop may join it in the onset. 

3. All remaining consonants are allocated to the coda of the preceding
syllable.

ENDUSAGE

####################################################################
# GLOBAL VARIABLES

# Math functions
sub abs { my($x)=shift; ($x>0) ? $x : -$x ; }
sub round { $_[0] =~ /([^\.]*)(.*)/; ($2 >= 0.5) ? $1+1 : $1; }
  
###################################################
# Subroutine to read an HTK lattice file
#
# Syntax: ($nodes, $edges, $header) = read_HTK_lattice(\*FILEHANDLE)
#
# $nodes is an arrayref of nodes
#    each node is an arrayref containing the following data:
#    [ back, forw, time ]
#    where forw is an arrayref of indices of forward edges,
#    back is an arrayref of indices of backward edges.
#    time is the time-alignment of this node
# $edges is an arrayref of edges
#    each edge, $e, is a hashref
#    tags on input edge are interpreted as key=val pairs,
#    and loaded as $$e{$key}=$val
#    For example, $$e{'W'} is set to the word label
#    
# $header is a hashref containing the header information
#
sub read_HTK_lattice {
  my($fh) = $_[0];
  my(@nodes) = ();
  my(@edges) = ();
  my(%header) = ();
  my($line);

  # Read lines of the file
  while($line = <$fh>) {
    # Node number given by I=
    if($line =~ /I=(\d+)/) {
      my($nodenum) = $1;
      unless(exists($nodes[$nodenum])) { $nodes[$nodenum] = [ [],[],0 ]; }
      if($line =~ /t=([\.\d]+)/) { $nodes[$nodenum][2] = $1; }
    }
    # Edge number given by J=
    elsif($line =~ s/J=(\d+)//) {
      my($edgenum) = $1;
      my(%edge)=();

      # Look for the startnode marker
      if($line =~ s/S=(\d+)//) { 
	$edge{'S'} = $1;
	# Tell the start node that I am one of its forward edges
	push(@{$nodes[$1][1]}, $edgenum);
      }
      
      # Look for the endnode marker
      if($line =~ s/E=(\d+)//) { 
	$edge{'E'} = $1; 
	# Tell the end node that I am one of its backward edges
	push(@{$nodes[$1][0]}, $edgenum);
      }
      
      # Read all other field-value pairs
      while($line =~ s/(\S+)=(\S+)//) { $edge{$1}=$2; }
      # Store the edge
      $edges[$edgenum] = \%edge;
    }
    # Otherwise, try to read header information
    else {
      while ($line =~ s/(\S+)=(\S+)//) { $header{$1} = $2; }
    }    
  }
  
  return(\@nodes, \@edges, \%header);
}

###################################################
# Subroutine to write an HTK lattice file
#
# Syntax: ($N, $L) = write_HTK_lattice($nodes, $edges, $header, \*FILEHANDLE)
#
# $nodes is an arrayref of nodes
# $edges is an arrayref of edges
# $header is a hashref containing HTK lattice header information
# FILEHANDLE is the opened output file
#
# $N is the number of nodes actually written to the file
# $L is the number of edges actually written to the file
#
sub write_HTK_lattice {
  my($nodes) = $_[0];
  my($edges) = $_[1];
  my($header) = $_[2];
  my($fh) = $_[3];
  my($N) = 0;
  my($L) = 0;

  # Print header information
  print $fh "VERSION=1.0\n";
  print $fh "UTTERANCE=",$$header{'UTTERANCE'},"\n";
  print $fh "lmname=",$$header{'lmname'},"\n";
  print $fh "lmscale=",$$header{'lmscale'}," ";
  print $fh "wdpenalty=",$$header{'wdpenalty'},"\n";
  print $fh "vocab=",$$header{'vocab'},"\n";
  print $fh sprintf("N=%-5d L=%-7d\n",0+@{$nodes},0+@{$edges});

  # Print nodes
  foreach $N (0..$#{$nodes}) {
    print $fh sprintf("I=%-5d ",$N);
    if (exists($$nodes[$N][2])) 
      { print $fh sprintf("t=%4.2f",$$nodes[$N][2]); }
    print $fh "\n";
  }
  # Print edges
  foreach $L (0..$#{$edges}) {
    my($e)=$$edges[$L];

    # Print the standard key-value pairs
    print $fh sprintf("J=%-7d S=%-5d E=%-5d W=%-20s ",$L, 
		      $$e{'S'}, $$e{'E'}, $$e{'W'});
    # Print all other key-value pairs
    foreach (keys %$e) {
      unless(/^[SEW]$/) { print $fh sprintf("%s=%s ", $_, $$e{$_}); }
    }
    # Print newline
    print $fh "\n";
  }
  return($N,$L);
}


###################################################
# Expand landmark-based transcription
#
# $aref = expand_transcription($string)
#
# $string contains colon-separated landmark specifications
# Each landmark specification is a set of comma-separated feature specs
# Each feature spec is a +, -, -+, or +- followed by the name of
#   the feature.
#
# $aref is an arrayref of hashrefs.
# $aref[$n] is the $nth expanded landmark specification
# $aref[$n]{'sonorant'} exists only if the feature 'sonorant'
#    was specified for the $nth landmark; if so, $aref[$n]{'sonorant'}
#    contains the value of the feature (-, +, -+, or +-).
#
sub expand_transcription {
  my(@lms) = split(/:/,$_[0]);
  my(@aref) = ();

  foreach $lm (@lms) {
    my(@feats) = split(/,/,$lm);
    my(%feattab) = ();
    
    # Value is the series of + and - specifiers at feature start,
    # Key is all remaining characters in the feature specification
    foreach $f (@feats) { 
       $f=~/([\+\-]+)(\S+)/; 
       $feattab{$2}=$1; 
     }
    
    push(@aref, \%feattab);
  }
  return(\@aref);
}
      
###################################################
# Syllabify a transcription
#
# $syls = syllabify_transcription($aref)
#
# $aref is an arrayref of hashrefs
# $$aref[$n]{'sonorant'} exists only if the feature 'sonorant'
#    was specified for the $nth landmark; if so, $aref[$n]{'sonorant'}
#    contains the value of the feature (-, +, -+, or +-).
#
# $syls is an arrayref
#   $$syls[$n][$m] is the index (in $aref) of the $mth landmark in the
#      $nth syllable
#
sub syllabify_transcription {
  my(@aref) = @{$_[0]};

  # Create the syllable arrays: one for each syllable nucleus
  my(@syls) = ();
  my($n) = 0;
  foreach $n (0..$#aref) 
    { if($aref[$n]{'syllabic'} eq '+') { push(@syls, [ $n ]); } }
  
  # Assign initial and final consonants
  $syls[0] =  [ 0..$syls[0][0] ];
  $syls[$#syls] = [ $syls[$#syls][0]..$#aref ];

  # Determine the place of articulation of each phoneme, if any
  #   (This is important for finding syllable affiliation)
  my(@place) = ();
  foreach $n (0..$#aref) {
    if($aref[$n]{'lips'} eq '+') { $place[$n]='lips'; }
    elsif($aref[$n]{'blade'} eq '+') { $place[$n]='blade'; }
    elsif($aref[$n]{'body'} eq '+') { $place[$n]='body'; }
  }

  # Assign internuclear consonants to coda or onset
  my($sylnum) = 0;
  foreach $sylnum (1..$#syls) {
    # Start and end of the intersyllabic region
    my($s) = $syls[$sylnum-1][$#{$syls[$sylnum-1]}]+1;
    my($e) = $syls[$sylnum][0]-1;
    # 1. Assign leftmost to coda, if it is a closure
    if(($aref[$s]{'sonorant'} eq '+-' || $aref[$s]{'continuant'} eq '+-') &&
       ($e >= $s)) { push(@{$syls[$sylnum-1]}, $s++); }
    
    # 2. Put everything possible into the onset
    # 2.a. Homorganic, similarly voiced stop-fric sequence = affricate
    if($aref[$e-1]{'continuant'} eq '-+' && $aref[$e]{'sonorant'} eq '-+' &&
       $place[$e-1] eq $place[$e] && 
       $aref[$e-1]{'voice'} eq $aref[$e]{'voice'} &&
       $e >= $s) { unshift(@{$syls[$sylnum]}, $e--); }
    # 2.b. Any one consonant release or -syllabic (glide) landmark
    if(($aref[$e]{'sonorant'} eq '-+' || $aref[$e]{'continuant'} eq '-+' ||
	$aref[$e]{'syllabic'} eq '-') && ($e >= $s))
      { unshift(@{$syls[$sylnum]}, $e--); }
    # 2.c Boundary between and /s/ and a voiceless stop
    if($aref[$e]{'continuant'} eq '+-' && $aref[$e]{'sonorant'} eq '-' &&
       $aref[$e]{'anterior'} eq '+' && $aref[$e]{'distributed'} eq '-' &&
       $aref[$e]{'voice'} eq '-' && $aref[$e+1]{'voice'} eq '-' &&
       $e >= $s) { unshift(@{$syls[$sylnum]}, $e--); }
    # 3. Assign all remaining intersyllabic landmarks to coda
    if($e >= $s) { push(@{$syls[$sylnum-1]}, $s..($e-1)); }
  }
  return(\@syls);
}

###################################################
# sprint a transcription
#
# $string = sprint_transcription($aref, $syls)
#
# $aref is an arrayref of hashrefs
# $$aref[$n]{'sonorant'} exists only if the feature 'sonorant'
#    was specified for the $nth landmark; if so, $aref[$n]{'sonorant'}
#    contains the value of the feature (-, +, -+, or +-).
# $syls is an arrayref
#   $$syls[$n][$m] is the index (in $aref) of the $mth landmark in the
#      $nth syllable
#
# $string is the printed string
#
sub sprint_transcription {
  my(@aref) = @{$_[0]};
  my(@syls) = @{$_[1]};
  my($string) = '';
  my($n); my($m);
  # Set interpolator to comma
  $"=',';

  foreach $n (0..$#syls) {
    # Before each syllable, if string is nonzero length, put a semicolon
    if(length($string)>0) { $string .= ';' };
    foreach $m (0..$#{$syls[$n]}) {
      # Get the landmark transcription
      $lm = $aref[$syls[$n][$m]];

      my(@feats) = ();

      # Print syllabic, sonorant, continuant first if they exist
      foreach $f ('syllabic', 'sonorant', 'continuant') {
	if(exists($$lm{$f})) { push(@feats, $$lm{$f}.$f) ; }
      }
      # Then print the other features
      foreach $f (keys %$lm) {
	if($f !~ /^(syllabic|sonorant|continuant)$/) {push(@feats,$$lm{$f}.$f);}
      }

      # Then concatenate them all to $string
      if(@feats > 0) { 
	if(length($string)>0) { $string .= ':' };
	$string .= "@feats";
      }
    }
  }
  # Reset interpolator, and return
  $"=' ';
  return($string);
}


##################################################
# Main program

if (@ARGV == 0) { print $USAGE; exit; }

# Read in the script file
$arg=shift;
open(F,$arg) || die "Unable to read from script file $arg: $!\n";
$script = ''; while (<F>) { $script .= $_; }
close(F);

# Main loop: process each pair of lattice files in the script
while($script =~ s/^\s*(\S+)\s+(\S+)//) {
  my($inlat) = $1; my($outlat) = $2;

  # Read the input lattice
  unless(open(LATFILE, $inlat)) {
    print STDERR "Unable to read from lattice $inlat: $!\n";
    next;
  }
  my(@graph) = read_HTK_lattice(\*LATFILE);
  close(LATFILE);

  my($edge);
  foreach $edge (@{$graph[1]}) {
    # Expand the pronunciation from a string to an array of hashrefs
    my($aref) = expand_transcription($$edge{'p'});
    # Determine syllable affiliation of each landmark
    my($syls) = syllabify_transcription($aref);
    # Print out the syllabified transcription
    $$edge{'p'} = sprint_transcription($aref, $syls);
  }
    
  # Write the modified lattice
  unless(open(LAT_OUT, ">$outlat")) {
    print STDERR "Unable to write lattice out to $outlat: $!\n";
    next;
  }
  write_HTK_lattice(@graph, \*LAT_OUT);
  close(LAT_OUT);

  # Write summary to STDERR
  $nn=0+@{$graph[0]};
  $ne=0+@{$graph[1]};
  print STDERR "$inlat -> $outlat, $nn nodes, $ne edges\n";
}
