#!/usr/bin/perl
# timit2swb: Mark Hasegawa-Johnson, April 2003
# Illinois Speech and Language Engineering (ISLE group)
#
# Usage text (printed if timit2swb.pl is called with no arguments):
$USAGE=<<ENDUSAGE;

timit2swb.pl: Convert TIMIT phoneme labels into ARPABET, similar to
the notation used for the ICSI transcriptions of Switchboard

Usage:
  timit2swb.pl input.mlf > output.mlf
Example: 
  timit2swb.pl TRAIN.mlf > TRAIN_swb.mlf

This script converts a TIMIT transcription into a Switchboard-style
ARPABET transcription.  Stop closures and releases are combined into a
single segment.  Unreleased stops are converted to the stop
character. Some nonstandard TIMIT phones are converted to similar
standard phones, possibly with underscore annotation (e.g., hh_v for
hv).

ENDUSAGE

####################################################################
# GLOBAL VARIABLES

# Conversion rates for various file formats
$MLFPERIOD = 0.0000001;

# Math functions
sub abs { my($x)=shift; ($x>0) ? $x : -$x ; }
sub round { $_[0] =~ /([^\.]*)(.*)/; ($2 >= 0.5) ? $1+1 : $1; }

###################################################
# Subroutine to read MLF-format transcription file
#
# Syntax: $data = read_MLF_file("filename");
#
# Returned value $data is a reference to a hash.
# Hash keys are the roots of filenames given in the MLF file.
# Hash values are references to phn-structure arrays, i.e.
#   arrays of the form returned by read_NIST_file or read_ALA_file:
#   Each element of the array is a subarray containing start time,
#   end time, and labels.
# $$data{$filename}[0][0] is starttime, $$data{$filename}[0][1] is endtime,
# @{$$data{$filename}[0]}[2..$#{$$data{$filename}[0]}] is the array
# slice containing all remaining labels 
#
sub read_MLF_file {
  # Initialize %MLF to null
  my(%MLF)  = ();

  # Open the INPUTFILE or die with an error message
  open(INPUTFILE,$_[0]) || die "Unable to open MLF file $_[0]: $!";

  # Read in all lines from the INPUTFILE
  while (<INPUTFILE>) {
    # If this line starts a transcription, read it
    if (/\".*\/([^\/\.]+)[^\/]*\"/) {
      my($filename) = $1;
      $MLF{$filename} = read_MLF_entry($filename, \*INPUTFILE);
    }
  }
  close(INPUTFILE);

  # Return a reference to the MLF structure
  return(\%MLF);
}
  
###################################################
# Subroutine to read MLF entry
#
# Syntax: $data = read_MLF_entry("filename", \*INPUTFILE);
#
# Returned value $data is an arrayref of arrayrefs
# $$tran[0][0] is first seg start time, $$tran[0][1] is end time,
# @$$tran[0][2..$#{$$tran[0]}] is the sequence of labels at first time.
#
sub read_MLF_entry {
  my($filename) = $_[0];
  my($fh) = $_[1];
  my(@tran) = ();
      
  # Read all consecutive legible transcription lines
  for($l=<$fh>; $l=~/(\d+)\s+(\d+)\s+(.*)/; $l=<$fh>) {
    my(@labels) = split(/\s+/,$3);
    # Push a reference to @record onto the @tran array
    push(@tran, [ $1 * $MLFPERIOD, $2 * $MLFPERIOD, @labels ] );
  }
  return(\@tran);
}

###################################################
# Print phoneme data to an MLF file
#
# Syntax: print_MLF_entry("root_filename", \@phndata, $filehandleref);
#
# The output filehandle MLF must be globally defined
# and already open.
#
sub print_MLF_entry {
  my($filename) = $_[0];
  my($phndata) = $_[1];
  my($fh) = $_[2];

  # Print the filename to the MLF
  $filename =~ s:.*/::g;
  $filename =~ s/\..*//g;
  print $fh "\"*/$filename\.lab\"\n";

  # Print out the converted data into MLF, and to the label file
  foreach $r (@$phndata) {
    my(@l) = @$r[2..$#$r];
    print $fh sprintf("%d %d ",round($$r[0]/$MLFPERIOD),
		      round($$r[1]/$MLFPERIOD)),"@l\n";
  }
  # Print final period that separates sentences
  print $fh '.',"\n";
}

###################################################
# Subroutine to convert TIMIT phones to SWB phone set
#
# Syntax: @data = timit2swb(@data)
# 
sub timit2swb {
  my($input) = $_[0];
  my(@data) = ();

  # Copy the phoneme records, so we don't modify the original
  foreach $r (@$input) { push(@data, [ @$r ]); }
  
  # Process each record, while there are records left
  for( my($i)=$#data; $i >= 0; $i-- ) {
    # Merge TIMIT labels as necessary to approximate Lee et al. 1990
    foreach $_ ( $data[$i][2] ) {
      /ax\-h/ && { $data[$i][2] = 'ax' };
      /axr/ && { $data[$i][2] = 'er' };
      /ux/ && { $data[$i][2] = 'uw' };
      /hv/ && { $data[$i][2] = 'hh_v' };
      /nx/ && { $data[$i][2] = 'n' };
      /(h\#)|(pau)/ && { $data[$i][2] = 'sil' } && { $data[$i][3] = 'sil' };
      /pcl/ && { $data[$i][2] = 'p' };
      /bcl/ && { $data[$i][2] = 'b' };
      /kcl/ && { $data[$i][2] = 'k' };
      /gcl/ && { $data[$i][2] = 'g' };
      /tcl/ && { $data[$i][2] = 't' };
      /dcl/ && { $data[$i][2] = 'd' };
    }
  }
  # Now delete stop closures iff following segment
  # is homorganic stop release.
  # Special case:
  #    vowel q becomes vowel_cr (creaky phone)
  #    q vowel becomes vowel_cr (creaky phone)
  #
  for( my($i)=$#data-1; $i >= 0; $i-- ) {
    # First, find necessary mergers: mark found mergers by setting 
    #   start times equal
    foreach $_ ( $data[$i][2] . ' ' . $data[$i+1][2] ) {
      /^t t$/ && { $data[$i+1][0]=$data[$i][0] };
      /^t ch$/ && { $data[$i+1][0]=$data[$i][0] };
      /^d jh$/ && { $data[$i+1][0]=$data[$i][0] };
      /^d d$/ && { $data[$i+1][0]=$data[$i][0] };
      /^p p$/ && { $data[$i+1][0]=$data[$i][0] };
      /^b b$/ && { $data[$i+1][0]=$data[$i][0] };
      /^k k$/ && { $data[$i+1][0]=$data[$i][0] };
      /^g g$/ && { $data[$i+1][0]=$data[$i][0] };
      /^[aeiou]. q$/ && { $data[$i+1][0]=$data[$i][0] };
      /^[aeiou]. q$/ && { $data[$i+1][2]=$data[$i][2] . '_cr' };
      /^q [aeiou].$/ && { $data[$i+1][0]=$data[$i][0] };
      /^q [aeiou].$/ && { $data[$i+1][2]=$data[$i+1][2] . '_cr' };
      /^epi / && { $data[$i+1][0]=$data[$i][0] };
    }
    # If segment to be eliminated includes higher levels,
    #   move them forward
    if ( $data[$i+1][0] == $data[$i][0] && @{$data[$i]} > 3 ) 
      { push( @{$data[$i+1]}, @{$data[$i]}[3..$#{$data[$i]}] ); }

    # Now delete the unnecessary segment
    ( $data[$i+1][0] == $data[$i][0] ) && splice(@data,$i,1);
  }

  # Return the modified @data array
  return(\@data);
} 

################################################################
# Main program

if (@ARGV==0) { print $USAGE; exit; }

foreach $MLF_file (@ARGV) {
  # Load the MLF
  $IMLF = read_MLF_file($MLF_file);

  # Convert each file to swb, in turn
  foreach $filename (keys %{$IMLF}) {
    # Convert the transcription
    $SMLF = timit2swb($$IMLF{$filename});

    # Print out the transcription to STDOUT
    print_MLF_entry($filename, $SMLF, \*STDOUT);
  }
}



